'''
Turtle experiment in tiny vector language
'''

import pygame
from pygame_ext import VectorSprite, Color, SimpleWorld, arc_through, draw_line
from pygame_events import EventListener, App
from geometry import rotate_on_origin_all, translate_all, point_at_distance, mag, angle_to_point

def mirror_x(points):
  temp = [(-x,y) for x,y in points]
  temp.reverse()
  return temp

def mirror_y(points):
  temp = [(x,-y) for x,y in points]
  temp.reverse()
  return temp

def mirror_xy(points):
  return [(-x,-y) for x,y in points]

def points(*coordinates, **kws):
  retVal = []
  dx, dy = kws.get('offset', (0,0))
  l = list(coordinates)
  while l:
    retVal.append((l.pop(0) + dx, l.pop(0) + dy))
  return retVal

class TurtleWorld(SimpleWorld):
  
  def __init__(self, surface):
    SimpleWorld.__init__(self, surface, None, (1,1), TurtleListener(self))
    self.background = pygame.Surface(surface.get_size(), 0, surface)
    self.background.fill(Color.white)
    self.sprites = {'turtle': Turtle(self.background)}
    self.current_sprite = self.sprites.values()[0]
    for sprite in self.sprites.values():
      sprite.draw(self.surface)
    pygame.display.update()

  def choose_sprite(self, pos):
    for sprite in self.sprites.values():
      if sprite.hit_test(pos):
        return sprite

  def select_sprite(self, sprite):
    self.current_sprite = sprite

  def turn_and_move(self, pos):
    sprite = self.current_sprite
    if not sprite: return
    sprite.turn_to_point(pos)
    sprite.move_to(pos)
    self.surface.blit(self.background, self.rect)
    sprite.draw(self.surface)
    pygame.display.update()

class TurtleListener(EventListener):

  def onpulse(self):
    #self.owner.draw()
    #self.owner.current_sprite.turn_to(all_angles.next())
    #self.owner.draw()
    pass

  def onclick(self, button, pos):
    # if click is on turtle, select turtle
    # if click is beginning of a drag, pen up
    sprite = self.owner.choose_sprite(pos)
    # if click is not on turtle, move to new location
    if sprite:
      self.owner.select_sprite(sprite)
    else:
      self.owner.turn_and_move(pos)
    

class Turtle(VectorSprite):

  def __init__(self, canvas):
    VectorSprite.__init__(self, (60,60))
    self.canvas = canvas
    self.angle = 0
    self.origin = self.rect.center
    self.render()
    self.is_pendown = True
    self.pen_color = Color.black
    self.pen_width = 1
    self.radius = 25 # used for hit test

  def hit_test(self, pos):
    return mag(self.rect.center, pos) < self.radius

  def render(self):
    outline = Color.black
    fill = Color.green
    dark_fill = Color.olivedrab
    highlight = Color.darkgreen
    VectorSprite.render(self)
    origin = x,y = self.origin
    angle = self.angle
    # head
    headPoints = points(15,2,18,2,18,4,20,4,25,2)
    headPoints += mirror_y(headPoints)
    headPoints = rotate_on_origin_all(headPoints, origin, angle)
    self.polygon(headPoints, fill_color=dark_fill)
    # tail
    tailPoints = points(-15,2,-25,0,-15,0) 
    tailPoints = rotate_on_origin_all(tailPoints, origin, angle)
    self.polygon(tailPoints, fill_color=dark_fill)
    # body
    bodyPoints = points(-15,4,-15,-4,-8,-10,8,-10,15,-4,15,4,8,10,-8,10) 
    bodyPoints = rotate_on_origin_all(bodyPoints, origin, angle)
    self.polygon(bodyPoints, fill_color=fill, stroke_color=outline)
    #inner shell
    innerPoints = points(-12,3,-12,-3,-6,-8,6,-8,13,-2,12,2,6,8,-6,8) 
    innerPoints = rotate_on_origin_all(innerPoints, origin, angle)
    self.polygon(innerPoints, fill_color=dark_fill)
    # shell markings
    curve1Points = points(-15,1,15,1,15,-1,-15,-1)
    curve1Points = rotate_on_origin_all(curve1Points, origin, angle)
    self.polygon(curve1Points, fill_color=highlight)
    curve2Points = points(-7,-10,-4,0,-7,10,-9,10,-6,0,-9,-10)
    curve2Points = rotate_on_origin_all(curve2Points, origin, angle)
    self.polygon(curve2Points,fill_color=highlight)
    linePoints = points(7,-10,4,0,7,10,9,10,6,0,9,-10)
    linePoints = rotate_on_origin_all(linePoints, origin, angle)
    self.polygon(linePoints,fill_color=highlight)
    # legs
    legPoints = points(0,0,3,3,5,3,5,4,7,4,7,0,5,0,5,1,4,1,2,-2)
    dx = 10
    dy = 10
    leg_1 = translate_all(legPoints, (dx, dy))
    leg_1 = rotate_on_origin_all(leg_1, origin, angle)
    leg_2 = translate_all(mirror_x(legPoints), (-dx, dy))
    leg_2 = rotate_on_origin_all(leg_2, origin, angle)
    leg_3 = translate_all(mirror_y(legPoints), (dx, -dy))
    leg_3 = rotate_on_origin_all(leg_3, origin, angle)
    leg_4 = translate_all(mirror_xy(legPoints), (-dx, -dy))
    leg_4 = rotate_on_origin_all(leg_4, origin, angle)
    self.polygon(leg_1, fill_color=dark_fill)
    self.polygon(leg_2, fill_color=dark_fill)
    self.polygon(leg_3, fill_color=dark_fill)
    self.polygon(leg_4, fill_color=dark_fill)

  def move_by(self, distance):
    origin = self.rect.center
    dest = point_at_distance(origin, self.angle, distance)
    self.move_to(dest)

  def move_to(self, dest):
    origin = self.rect.center
    self.rect.center = dest
    if self.is_pendown:
      pygame.draw.line(self.canvas, self.pen_color, origin, dest, 1)

  def turn_by(self, angle):
    self.angle = (self.angle +  angle) % 360
    self.render()

  def turn_to_point(self, pt):
    origin = self.rect.center
    angle = angle_to_point(origin, pt)
    print 'turn to ', angle
    self.turn_to(angle)

  def turn_to(self, angle):
    self.angle = angle
    self.render()

def main():
  app = App()
  app.add_world(TurtleWorld(app.surface))
  app.run()

if __name__ == '__main__':
  main()

