'''
  A game based on Tantrix
'''

import pygame
from geometry import *
from pygame_ext import VectorSprite, Color, draw_arc
from pygame_events import App, EventDispatcher, EventListener

GAP_RADIUS = 1.05

class HexWorld:

  def __init__(self, surface):
    self.surface = surface
    self.rect = self.surface.get_rect()
    self.size = self.width, self.height = self.rect.size
    self.display_area = self.rect.size
    self.listener = BoardListener(self)
    self.center = self.width * 0.5, self.height * 0.5
    pygame.draw.rect(self.surface, Color.white, self.rect, 0)
    screenradius = min(self.center[X], self.center[Y])
    boardradius = screenradius * 0.8
    board_points = hexpoints(self.center, screenradius, True)
    board_colors = [Color.red, Color.blue, Color.yellow] * 2
    for i in range(6):
      pygame.draw.polygon(self.surface, board_colors[i], (board_points[i-1], board_points[i], self.center))
    pygame.draw.polygon(self.surface, Color.white, hexpoints(self.center,
    screenradius * 0.9, True))
    self.hexradius = (boardradius / 7.0)
    self.centers = hexcenters(self.center, self.hexradius * GAP_RADIUS)
    self.hexen = [Hex(c, self.hexradius, rndPath()) for c in self.centers]
    #pygame.display.update()
    for hex in self.hexen: hex.draw(self.surface)
    # is the next line a good idea?
    pygame.display.update()

class BoardListener(EventListener):

  def onclick(self, button, pos):
    for hex in self.owner.hexen:
      if hex.hittest(pos):
        hex.onclick(button, pos)
        hex.draw(self.owner.surface)
        pygame.display.update(hex.rect)
        return

  def onmousemove(self, pos, prev):
    curr = None
    prior = None
    for hex in self.owner.hexen:
      if hex.hittest(pos):
        curr = hex
      if hex.hittest(prev):
        prior = hex
    if curr == prior:
      if curr:
        curr.onmousemove(pos, prev)
    else:
      if prior:
        prior.onmouseout(pos)
      if curr:
        curr.onmouseover(pos)
        curr.onmousemove(pos, prev)

class Hex(VectorSprite):
  def __init__(self, center, radius, paths, bg=Color.black, fg=Color.grey):
    VectorSprite.__init__(self)
    self.center = center
    #print 'center:', center, ', radius: ', radius
    self.radius = radius
    self.points = hexpoints(center, radius)
    self.midpoints = [midpoint(self.points[i], self.points[i-1]) for i in hexsides]
    # self.rect is the bounding box
    b_x, b_y = self.points[3][0], self.points[4][1]
    b_width, b_height = self.points[0][0] - b_x, self.points[1][1] - b_y
    self.rect = pygame.Rect(b_x, b_y, b_width, b_height)
    # self hitrect is bounding box for hittest (naive implementation)
    h_x, h_y = self.midpoints[3][0], b_y
    h_width, h_height = self.midpoints[1][0] - h_x, b_height
    self.hitrect = pygame.Rect(h_x, h_y, h_width, h_height)
    self.borderwidth = int(radius * 0.05)
    self.fg = fg
    self.bg = bg
    self.paths = paths
    self.red = paths[0:2]
    self.blue = paths[2:4]
    self.yellow = paths[4:6]
    self.resize(center, radius)

  def hittest(self, point):
    return self.hitrect.collidepoint(*point)

  def onclick(self, button, pos):
    if pos[0] < self.center[0]:
      self.rotate_left()
    else:
      self.rotate_right()

  def onmousemove(self, pos, prev):
    if pos[0] < self.center[0]:
      pygame.mouse.set_cursor(*pygame.cursors.tri_left)
    else:
      pygame.mouse.set_cursor(*pygame.cursors.tri_right)

  def onmouseover(self, pos):
    pass

  def onmouseout(self, pos):
    pygame.mouse.set_cursor(*pygame.cursors.arrow)

  def rotate_left(self):
    self.paths = [(i - 1) % 6 for i in  self.paths]
    self.red = self.paths[0:2]
    self.blue = self.paths[2:4]
    self.yellow = self.paths[4:6]

  def rotate_right(self):
    self.paths = [(i + 1) % 6 for i in  self.paths]
    self.red = self.paths[0:2]
    self.blue = self.paths[2:4]
    self.yellow = self.paths[4:6]

  def resize(self, center, radius):
    self.center = center
    self.radius = radius
    self.points = hexpoints(self.center, radius)
    self.midpoints = [midpoint(self.points[i], self.points[i-1]) for i in hexsides] # midpoints of hex edges
    self.borderwidth = radius * 0.05
    #self.polygon = polygon(self.points, self.borderwidth)

  def draw(self, board):
    self.board = board
    pygame.draw.polygon(board, self.bg, self.points, 0)
    # check paths of various colors
    self.curveWidth = self.radius * 0.14
    # find a point about 1/4 of the way from the centerpoint, off the midpoint line
    def p(p1, p2): return midpoint(midpoint(midpoint(p1,p2),p2),self.center)
    # Move points away from each other, keeping them on the same line
    def c(t): return ex(p(t[1], t[0]), p(t[0], t[1])) # controls for curves
    def m(i,j): return self.midpoints[i], self.midpoints[j] # shorthand 
    if self.red: self.curve(self.red[0], self.red[1], Color.red)
    if self.blue: self.curve(self.blue[0], self.blue[1], Color.blue)
    if self.yellow: self.curve(self.yellow[0], self.yellow[1], Color.yellow)
    #for i in hexsides:
    #  pos = translate(self.center, polarToCart(self.radius * GAP_RADIUS, PI_3RD * i))
    #  pygame.draw.circle(board, Color.green, pos, 1, 0)
    #  colors = [Color.red, Color.green, Color.blue, Color.yellow, Color.plum, Color.orange]
    #  m = mag(self.center, self.midpoints[i])
    #  pos2 = translate(self.center, polarToCart(m * GAP_RADIUS * 2, PI_3RD * i + PI_6TH))
    #  pygame.draw.circle(board, colors[i], pos2, 6 - i, 0)
    pygame.draw.aalines(board, self.fg, 1, self.points, 1)

  def curve(self, s1, s2, color):
    dS = abs(s1 - s2)
    if dS == 1 or dS == 5:
      self.small_arc(s1, s2, color)
    elif dS == 2 or dS == 4:
      self.large_arc(s1, s2, color)
    elif dS == 3:
      self.line(s1, s2, color)
    else:
      print 'error in curve:', s1, s2

  def small_arc(self, s1, s2, color):
    index = sm_index(s1, s2)
    center_angle = PI_3RD * index
    #print 'center index: %i, rads: %f, degrees: %f' % (index, center_angle/pi, radToDegree(center_angle))
    offset = self.radius * GAP_RADIUS 
    center = translate(self.center, polarToCart(offset, center_angle))
    start_angle, stop_angle = small_angles[index]
    #stop_angle = start_angle + PI_3RD
    #pygame.draw.circle(self.board, Color.green, center, 1)
    #pygame.draw.rect(self.board, Color.yellow, arc_rect, 1)
    #print 'small %i: start %f, end %f' % (index ,start_angle/pi,stop_angle/pi)
    draw_arc(self.board, color, center, offset * 0.5, start_angle, stop_angle, self.curveWidth)

  def large_arc(self, s1, s2, color):
    side_index = lg_index(s1, s2)
    center_angle = PI_3RD * side_index + PI_6TH
    m = mag(self.center, self.midpoints[side_index])
    offset = m * GAP_RADIUS * 2
    offset_angle = PI_3RD * side_index + PI_6TH
    center = translate(self.center, polarToCart(offset, offset_angle))
    start_angle, stop_angle = large_angles[side_index]
    draw_arc(self.board, color, center, offset, start_angle, stop_angle, self.curveWidth)

  def line(self, s1, s2, color):
    pygame.draw.line(self.board, color, self.midpoints[s1], self.midpoints[s2], self.curveWidth)

  def hitTest(self, pt):
    return self.polygon.containsPoint_(pt)

  def dodecant(self, pt):
    localPt = pt[0] - self.center[0], pt[1] - self.center[1]
    r, theta = cartToPolar(localPt[0], localPt[1])
    return theta

def sm_index(a,b):
  if a == 5 and b == 0: return 5
  if a == 0 and b == 5: return 5
  return min(a,b)

def lg_index(a,b):
  if a == 5 and b == 1: return 5
  if a == 1 and b == 5: return 5
  if a == 4 and b == 0: return 4
  if a == 0 and b == 4: return 4
  return min(a,b)

large_angles = (
  (2 * PI_3RD, pi),
  (PI_3RD, 2 * PI_3RD),
  (0, PI_3RD),
  (5 * PI_3RD, 2 * pi),
  (4 * PI_3RD, 5 * PI_3RD),
  (pi, 4 * PI_3RD),
)

small_angles = (
  (2*PI_3RD,4*PI_3RD),
  (PI_3RD, 3*PI_3RD),
  (0,2*PI_3RD),
  (5*PI_3RD,PI_3RD),
  (4*PI_3RD,2*pi),
  (pi,5*PI_3RD),
  )

def rndPath():
  #return [0,2,1,5,3,4]
  import random
  paths = range(6)
  random.shuffle(paths)
  return paths

def hexcenters(center, r):
  d =  1.5 * r
  w = 0.5 * W * r
  return translateAll(((0*d,8*w),
          (-1*d,7*w),(1*d,7*w),
          (-2*d,6*w),(0*d,6*w),(2*d,6*w),
          (-3*d,5*w),(-1*d,5*w),(1*d,5*w),(3*d,5*w),
          (-4*d,4*w),(-2*d,4*w),(0*d,4*w),(2*d,4*w),(4*d,4*w),
          (-3*d,3*w),(-1*d,3*w),(1*d,3*w),(3*d,3*w),
          (-4*d,2*w),(-2*d,2*w),(0*d,2*w),(2*d,2*w),(4*d,2*w),
          (-3*d,1*w),(-1*d,1*w),(1*d,1*w),(3*d,1*w),
          (-4*d,0*w),(-2*d,0*w),(0*d,0*w),(2*d,0*w),(4*d,0*w),
          (-3*d,-1*w),(-1*d,-1*w),(1*d,-1*w),(3*d,-1*w),
          (-4*d,-2*w),(-2*d,-2*w),(0*d,-2*w),(2*d,-2*w),(4*d,-2*w),
          (-3*d,-3*w),(-1*d,-3*w),(1*d,-3*w),(3*d,-3*w),
          (-4*d,-4*w),(-2*d,-4*w),(0*d,-4*w),(2*d,-4*w),(4*d,-4*w),
          (-3*d,-5*w),(-1*d,-5*w),(1*d,-5*w),(3*d,-5*w),
          (-2*d,-6*w),(0*d,-6*w),(2*d,-6*w),
          (-1*d,-7*w),(1*d,-7*w),
          (0*d,-8*w)), center)


def main():
  app = App()
  world = HexWorld(app.screen)
  app.add_world(world)
  app.run()

if __name__ == '__main__':
  main()

